/***********************************************************************************
* Smooth Tasks
* Copyright (C) 2009 Marcin Baszczewski <marcin.baszczewski@gmail.com>
*                    Mathias panzenböck <grosser.meister.morti@gmx.net>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

// Smooth Tasks
#include "SmoothTasks/Light.h"
#include "SmoothTasks/Applet.h"
#include "SmoothTasks/TaskItem.h"

// Qt
#include <QTimer>
#include <QPainter>
#include <QStyleOptionGraphicsItem>

// Plasma
#include <Plasma/Applet>
#include <Plasma/PaintUtils>

// KDE
#include <KIcon>
#include <KIconEffect>
#include <KIconLoader>

// Other
#include <cmath>

namespace SmoothTasks {

Light::Light(TaskItem *item) : QObject(item),
	m_item(item),
	m_count(0),
	m_currentAnimationDuration(0),
	m_animation(0),
	m_progress(0.0),
	m_move(true),
	m_currentAnimation(NoAnimation),
	m_animationRepeater(new QTimer()),
	m_repeater(true),
	m_point(),
	m_mouse(false) {
}

Light::~Light() {
	if (m_animation) {
		Plasma::Animator::self()->stopCustomAnimation(m_animation);
	}
	m_animationRepeater->deleteLater();
}

void Light::paint(QPainter *p, const QRectF& boundingRect, const bool isRotated) {
	if (!m_mouse && !m_animation) {
		return;
	}

	QRectF drawRect(boundingRect);
	QRadialGradient radialGrad;

	drawRect.setLeft(drawRect.left()     - 4);
	drawRect.setTop(drawRect.top()       - 4);
	drawRect.setWidth(drawRect.width()   + 4);
	drawRect.setHeight(drawRect.height() + 4);

	switch (m_currentAnimation) {
	case StartupAnimation:   setupStartupAnimation(drawRect, radialGrad);       break;
	case AttentionAnimation: setupAttentionAnimation(drawRect, radialGrad);     break;
	case NoAnimation:        setupNoAnimation(drawRect, radialGrad, isRotated); break;
	default: return;
	}
	
	QBrush oldBrush(p->brush());
	QPen   oldPen(p->pen());

	p->setBrush(radialGrad);
	p->setPen(Qt::NoPen);
	p->drawRect(drawRect);
	p->setBrush(oldBrush);
	p->setPen(oldPen);
}

void Light::setupStartupAnimation(QRectF& drawRect, QRadialGradient& radialGrad) {
	QPointF pos(drawRect.width() * 0.5, drawRect.height() * 0.5);
	qreal msize = drawRect.width();
	
	QColor lightColor(m_item->applet()->lightColor());
	lightColor.setAlpha(200);

	radialGrad = QRadialGradient(pos, msize * 2 *
		(m_progress < 0.5 ? (m_progress * 0.5 + 0.5) : (1 - m_progress / 2)));
	radialGrad.setColorAt(0, lightColor);
	radialGrad.setColorAt(0.5, QColor(0, 0, 0, 0));
}

void Light::setupAttentionAnimation(QRectF& drawRect, QRadialGradient& radialGrad) {
	QPointF pos(drawRect.width() * 0.5, drawRect.height() * 0.5);
	qreal msize = drawRect.width();

	QColor lightColor(m_item->applet()->lightColor());
	lightColor.setAlpha(200);

	radialGrad = QRadialGradient(pos, msize * 2 *
		(m_progress < 0.5 ? (m_progress * 0.5 + 0.5) : (1 - m_progress / 2)));
	radialGrad.setColorAt(0, lightColor);
	radialGrad.setColorAt(0.5, QColor(0, 0, 0, 0));
}

void Light::setupNoAnimation(QRectF& drawRect, QRadialGradient& radialGrad, const bool isRotated) {
	QPointF pos;
	qreal msize, R, grad, dist;

	qreal width  = drawRect.width();
	qreal height = drawRect.height();
	
	R     = (height * height + width * width) / (4 * height); // radius of circle
	msize = R + height / 4;                                   // real visible radius
	grad  = R > height ? (R - height / 1.5) / R : 0;          // fit gradient to radius
	dist  = R > height ?  R + height / 2 : height;            // distance of centre of circle from widget

	if (isRotated) {
		pos.setX(width - m_point.y());
		pos.setY(dist);
	}
	else {
		pos.setX(m_point.x());
		pos.setY(dist);
	}

	QColor lightColor(m_item->applet()->lightColor());
	lightColor.setAlpha(200);

	radialGrad = QRadialGradient(pos, msize);
	radialGrad.setColorAt(grad, lightColor);
	radialGrad.setColorAt(1, QColor(0, 0, 0, 0));
}

void Light::setPoint(const QPointF& point) {
	m_point = point;
}

void Light::setMouse(bool mouse) {
	m_mouse = mouse;
}

void Light::startAnimation(AnimationType animation, int duration, bool repeater) {
	m_count = 0;
	m_currentAnimationDuration = duration;
	m_currentAnimation = animation;

	m_animationRepeater->setInterval(duration);
	m_animationRepeater->start();
	
	repeatAnimation();

	m_repeater = repeater;
	connect(m_animationRepeater, SIGNAL(timeout()), this, SLOT(repeatAnimation()));
}

void Light::stopAnimation() {
	m_animationRepeater->stop();
	m_currentAnimation = NoAnimation;
	Plasma::Animator::self()->stopCustomAnimation(m_animation);
	m_animation = 0;
}

void Light::repeatAnimation() {	
	if (!m_repeater && m_count != 0) {
		return;
	}

	if (m_animation) {
		Plasma::Animator::self()->stopCustomAnimation(m_animation);
		m_animation = 0;
	}
	int frames = m_item->applet()->fps() * m_currentAnimationDuration / 1000;

	if (frames <= 0) {
		animation(1.0);
	}
	else {
		m_animation = Plasma::Animator::self()->customAnimation(
			frames, m_currentAnimationDuration,
			Plasma::Animator::LinearCurve,
			this, "animation");
	}
	++ m_count;
}

void Light::animation(qreal progress) {
	m_progress = progress;  
	emit update();
}

} // namespace SmoothTasks
#include "Light.moc"
