/***********************************************************************************
* Smooth Tasks
* Copyright (C) 2009 Marcin Baszczewski <marcin.baszczewski@gmail.com>
* Copyright (C) 2009 Mathias Panzenböck <grosser.meister.morti@gmx.net>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*
***********************************************************************************/

// Smooth Tasks
#include "SmoothTasks/TaskIcon.h"
#include "SmoothTasks/TaskItem.h"
#include "SmoothTasks/Applet.h"

// Qt
#include <QTimer>
#include <QPainter>
#include <QStyleOptionGraphicsItem>
#include <QFont>
#include <QApplication>

// Plasma
#include <Plasma/Applet>
#include <Plasma/PaintUtils>

// KDE
#include <KIcon>
#include <KIconEffect>
#include <KIconLoader>

namespace SmoothTasks {

TaskIcon::TaskIcon(TaskItem *item)
	: QObject(item),
	 m_item(item),
	 m_icon(),
	 m_rect(),
	 m_pixmap(),
	 m_animationRepeater(new QTimer(this)),
	 m_currentAnimationDuration(0),
	 m_animation(0),
	 m_progress(0.0),
	 m_type(Other) {
}

TaskIcon::~TaskIcon() {
	if (m_animation) {
		Plasma::Animator::self()->stopCustomAnimation(m_animation);
	}
}

void TaskIcon::updatePos() {
	qreal size     = this->size();
	QSize iconSize = m_icon.actualSize(QSize(size, size));
	QRectF boundingRect;
	const QSizeF& cellSize(m_item->cellSize());
	const bool isVertical = m_item->orientation() == Qt::Vertical;

	if (isVertical) {
		boundingRect = QRectF(
			m_rect.left(),
			qMax(m_rect.top(), m_rect.bottom() - cellSize.width()),
			qMin(cellSize.height(), m_rect.width()),
			qMin(cellSize.width(),  m_rect.height()));
	}
	else {
		boundingRect = QRectF(m_rect.left(), m_rect.top(),
			qMin(cellSize.width(),  m_rect.width()),
			qMin(cellSize.height(), m_rect.height()));
	}

	QPointF center(boundingRect.center());
	m_pos.setX(center.x() - (iconSize.width()  * 0.5));
	m_pos.setY(center.y() - (iconSize.height() * 0.5));

	if (QApplication::isRightToLeft()) {
		if (isVertical) {
			m_pos.setY(m_rect.height() - m_pos.y() - iconSize.height());
		}
		else {
			m_pos.setX(m_rect.width() - m_pos.x() - iconSize.width());
		}
	}
}

void TaskIcon::paint(QPainter *p, qreal hover) {
	m_pixmap = m_icon.pixmap(size());

	if (m_pixmap.isNull()) {
		kDebug() << "TaskIcon pixmap is null";
		return;
	}

	if (m_animation) {
		animationStartup(m_progress);
	}
	
	if (hover > 0.0) {
		animationHover(hover);
	}

	if (type() == TaskIcon::Group) {
		QPainter iconPainter(&m_pixmap);
		QPixmap  icon(KIcon("document-multiple").pixmap(
			m_pixmap.width()  * 0.45,
			m_pixmap.height() * 0.45));

		iconPainter.drawPixmap(
			m_pixmap.width()  - icon.width(),
			m_pixmap.height() - icon.height(),
			icon);
		
		iconPainter.end();
	}
	p->drawPixmap(m_pos, m_pixmap);
}

void TaskIcon::setRect(QRectF rect) {
	if (rect != m_rect) {
		m_rect = rect;
		updateIcon(m_icon);
	}
}

void TaskIcon::setIcon(QIcon icon) {
	updateIcon(icon);
}

void TaskIcon::setType(IconType type) {
	m_type = type;
}

TaskIcon::IconType TaskIcon::type() const {
	return m_type;
}

void TaskIcon::updateIcon(QIcon icon) {
	if (!icon.isNull()) {
		m_icon = icon;
	}
	updatePos();
}

qreal TaskIcon::size() const {
	qreal size = qMin(m_rect.width(), m_rect.height()) * m_item->applet()->iconScale();

	if (size < 1.0) {
		size = 1.0;
	}
	
	return size;
}

void TaskIcon::startStartupAnimation(int duration) {
	m_currentAnimationDuration = duration;

	m_animationRepeater->setInterval(duration);
	m_animationRepeater->start();
	
	repeatAnimation();

	connect(m_animationRepeater, SIGNAL(timeout()), this, SLOT(repeatAnimation()));
}

void TaskIcon::stopStartupAnimation() {
	m_animationRepeater->stop();

	if (m_animation) {
		Plasma::Animator::self()->stopCustomAnimation(m_animation);
		m_animation = 0;
	}
}

void TaskIcon::repeatAnimation() {
	if (m_animation) {
		Plasma::Animator::self()->stopCustomAnimation(m_animation);
	}

	m_animation = Plasma::Animator::self()->customAnimation(
		m_item->applet()->fps() * m_currentAnimationDuration / 1000,
		m_currentAnimationDuration,
		Plasma::Animator::LinearCurve,
		this, "animation");
}

void TaskIcon::animation(qreal progress) {
	m_progress = progress;  
	emit update();
}

void TaskIcon::animationHover(qreal hover) {
	KIconEffect *effect = KIconLoader::global()->iconEffect();
	
	if (effect->hasEffect(KIconLoader::Desktop, KIconLoader::ActiveState)) {
		if (qFuzzyCompare(qreal(1.0), hover)) {
			m_pixmap = effect->apply(
				m_pixmap,
				KIconLoader::Desktop,
				KIconLoader::ActiveState);
		}
		else if (hover != 0.0) {
			m_pixmap = Plasma::PaintUtils::transition(
				m_pixmap,
				effect->apply(
					m_pixmap,
					KIconLoader::Desktop,
					KIconLoader::ActiveState),
				hover);
		}
	}
}

void TaskIcon::animationStartup(qreal progress) {
	QPixmap pixmap = QPixmap(m_pixmap.width(), m_pixmap.height());
	pixmap.fill(Qt::transparent);
	int width;
	int height;

	if (progress < 0.5) {
		width  = m_pixmap.width()  * (0.5 + progress * 0.5);
		height = m_pixmap.height() * (1.0 - progress * 0.5);
	}
	else {
		width  = m_pixmap.width()  * (1.0 - progress * 0.5);
		height = m_pixmap.height() * (0.5 + progress * 0.5);
	}

	QPixmap scaled = m_pixmap.scaled(
		width, height,
		Qt::IgnoreAspectRatio,
		Qt::SmoothTransformation);
	
	if (!scaled.isNull()) {
		QPainter pixmapPainter(&pixmap);
		pixmapPainter.drawPixmap(
			(m_pixmap.width()  - width)  / 2,
			(m_pixmap.height() - height) / 2,
			scaled);


		pixmapPainter.end();
	}
	m_pixmap = pixmap;

	QPixmap transparentPixmap = QPixmap(m_pixmap.width(), m_pixmap.height());
	transparentPixmap.fill(Qt::transparent);
	m_pixmap = Plasma::PaintUtils::transition(transparentPixmap, m_pixmap, 0.85);
}

} // namespace SmoothTasks
#include "TaskIcon.moc"
